@@ -1,3 +1,22 @@
+# This file is part of the TREZOR project.
+#
+# Copyright (C) 2012-2016 Marek Palatinus <slush@satoshilabs.com>
+# Copyright (C) 2012-2016 Pavol Rusnak <stick@satoshilabs.com>
+# Copyright (C) 2016      Jochen Hoenicke <hoenicke@gmail.com>
+#
+# This library is free software: you can redistribute it and/or modify
+# it under the terms of the GNU Lesser General Public License as published by
+# the Free Software Foundation, either version 3 of the License, or
+# (at your option) any later version.
+#
+# This library is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU Lesser General Public License for more details.
+#
+# You should have received a copy of the GNU Lesser General Public License
+# along with this library.  If not, see <http://www.gnu.org/licenses/>.
+
 from __future__ import print_function, absolute_import
 
 import os
@@ -30,6 +49,29 @@ DEFAULT_CURVE = 'secp256k1'
 # monkeypatching: text formatting of protobuf messages
 tools.monkeypatch_google_protobuf_text_format()
 
+
+def getch():
+    try:
+        import termios
+    except ImportError:
+        # Non-POSIX. Return msvcrt's (Windows') getch.
+        import msvcrt
+        return msvcrt.getch()
+
+    # POSIX system. Create and return a getch that manipulates the tty.
+    import sys, tty
+    def _getch():
+        fd = sys.stdin.fileno()
+        old_settings = termios.tcgetattr(fd)
+        try:
+            tty.setraw(fd)
+            ch = sys.stdin.read(1)
+        finally:
+            termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
+        return ch
+
+    return _getch()
+
 def get_buttonrequest_value(code):
     # Converts integer code to its string representation of ButtonRequestType
     return [ k for k, v in types.ButtonRequestType.items() if v == code][0]
@@ -174,6 +216,29 @@ class TextUIMixin(object):
         # log("Sending ButtonAck for %s " % get_buttonrequest_value(msg.code))
         return proto.ButtonAck()
 
+    def callback_RecoveryMatrix(self, msg):
+        if self.recovery_matrix_first_pass:
+            self.recovery_matrix_first_pass = False
+            log("Use the numeric keypad to describe positions.  For the word list use only left and right keys. The layout is:")
+            log("    7 8 9     7 | 9")
+            log("    4 5 6     4 | 6")
+            log("    1 2 3     1 | 3")
+        while True:
+            character = getch()
+            if character in ('\x03', '\x04'):
+                return proto.Cancel()
+
+            if character in ('\x08', '\x7f'):
+                return proto.WordAck(word='\x08')
+
+            # ignore middle column if only 6 keys requested.
+            if (msg.type == types.WordRequestType_Matrix6 and
+                character in ('2', '5', '8')):
+                continue
+
+            if (ord(character) >= ord('1') and ord(character) <= ord('9')):
+                return proto.WordAck(word=character)
+
     def callback_PinMatrixRequest(self, msg):
         if msg.type == 1:
             desc = 'current PIN'
@@ -204,6 +269,9 @@ class TextUIMixin(object):
             exit()
 
     def callback_WordRequest(self, msg):
+        if msg.type in (types.WordRequestType_Matrix9,
+                        types.WordRequestType_Matrix6):
+            return self.callback_RecoveryMatrix(msg)
         log("Enter one word of mnemonic: ")
         try:
             word = raw_input()
@@ -354,7 +422,7 @@ class DebugLinkMixin(object):
 
 class ProtocolMixin(object):
     PRIME_DERIVATION_FLAG = 0x80000000
-    VENDORS = ('bitcointrezor.com',)
+    VENDORS = ('bitcointrezor.com', 'trezor.io')
 
     def __init__(self, *args, **kwargs):
         super(ProtocolMixin, self).__init__(*args, **kwargs)
@@ -384,6 +452,16 @@ class ProtocolMixin(object):
             return []
 
         n = n.split('/')
+
+        # m/a/b/c => a/b/c
+        if n[0] == 'm':
+            n = n[1:]
+
+        # coin_name/a/b/c => 44'/SLIP44_constant'/a/b/c
+        coins = { "Bitcoin": 0, "Testnet": 1, "Namecoin": 7, "Litecoin": 2, "Dogecoin": 3, "Dash": 5, "Zcash": 133, }
+        if n[0] in coins:
+            n = ["44'", "%d'" % coins[n[0]] ] + n[1:]
+
         path = []
         for x in n:
             prime = False
@@ -411,12 +489,12 @@ class ProtocolMixin(object):
 
     @field('address')
     @expect(proto.Address)
-    def get_address(self, coin_name, n, show_display=False, multisig=None):
+    def get_address(self, coin_name, n, show_display=False, multisig=None, script_type=types.SPENDADDRESS):
         n = self._convert_prime(n)
         if multisig:
-            return self.call(proto.GetAddress(address_n=n, coin_name=coin_name, show_display=show_display, multisig=multisig))
+            return self.call(proto.GetAddress(address_n=n, coin_name=coin_name, show_display=show_display, multisig=multisig, script_type=script_type))
         else:
-            return self.call(proto.GetAddress(address_n=n, coin_name=coin_name, show_display=show_display))
+            return self.call(proto.GetAddress(address_n=n, coin_name=coin_name, show_display=show_display, script_type=script_type))
 
     @field('address')
     @expect(proto.EthereumAddress)
@@ -425,7 +503,7 @@ class ProtocolMixin(object):
         return self.call(proto.EthereumGetAddress(address_n=n, show_display=show_display))
 
     @session
-    def ethereum_sign_tx(self, n, nonce, gas_price, gas_limit, to, value, data=None):
+    def ethereum_sign_tx(self, n, nonce, gas_price, gas_limit, to, value, data=None, chain_id=None):
         def int_to_big_endian(value):
             import rlp.utils
             if value == 0:
@@ -449,6 +527,9 @@ class ProtocolMixin(object):
             data, chunk = data[1024:], data[:1024]
             msg.data_initial_chunk = chunk
 
+        if chain_id:
+            msg.chain_id = chain_id
+
         response = self.call(msg)
 
         while response.HasField('data_length'):
@@ -731,19 +812,21 @@ class ProtocolMixin(object):
 
     @field('message')
     @expect(proto.Success)
-    def recovery_device(self, word_count, passphrase_protection, pin_protection, label, language):
+    def recovery_device(self, word_count, passphrase_protection, pin_protection, label, language, type=types.RecoveryDeviceType_ScrambledWords):
         if self.features.initialized:
             raise Exception("Device is initialized already. Call wipe_device() and try again.")
 
         if word_count not in (12, 18, 24):
             raise Exception("Invalid word count. Use 12/18/24")
 
+        self.recovery_matrix_first_pass = True
         res = self.call(proto.RecoveryDevice(word_count=int(word_count),
                                    passphrase_protection=bool(passphrase_protection),
                                    pin_protection=bool(pin_protection),
                                    label=label,
                                    language=language,
-                                   enforce_wordlist=True))
+                                   enforce_wordlist=True,
+                                   type=type))
 
         self.init_device()
         return res
